import numpy as np
import pandas as pd
from discretize import TensorMesh
from discretize.utils import volume_average
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
from scipy.spatial import cKDTree as kdtree

from .datasets import EMDataset
from .utils import inverse_distance_interpolation


def find_locations_in_distance(xy_input, xy_output, distance=100.0):
    """
    Find indicies of locations of xy_output within a given separation distance
    from locations of xy_input.

    Parameters
    ----------

    xy_input: (*,2) array_like
        Input locations.
    xy_output: (*,2) array_like
        Ouput Locations where the indicies of the locations are sought.
    distance: float
        Separation distance used as a threshold

    Returns
    -------
    pts : (*,2) ndarray, float
        Sought locations.
    inds: (*,) ndarray, integer
        Sought indicies.
    """
    tree = kdtree(xy_output)
    out = tree.query_ball_point(xy_input, distance)
    temp = np.unique(out)
    inds = []
    for ind in temp:
        if ind != []:
            inds.append(ind)
    if len(inds) == 0:
        return None, None
    inds = np.unique(np.hstack(inds))
    pts = xy_output[inds, :]
    return pts, inds


def find_closest_locations(xy_input, xy_output):
    """
    Find indicies of the closest locations of xy_output from from locations of xy_input.

    Parameters
    ----------

    xy_input: (*,2) array_like
        Input locations.
    xy_output: (*,2) array_like
        Ouput Locations where the indicies of the locations are sought.

    Returns
    -------
    d : (*,) ndarray, float
        Closest distance.
    inds: (*,) ndarray, integer
        Sought indicies.
    """
    tree = kdtree(xy_output)
    d, inds = tree.query(xy_input)
    return d, inds


def compute_fraction_for_aem_layer(hz, lith_data):
    """
    Compute fraction of lithology in AEM layers

    Parameters
    ----------

    hz: (n_layer,) array_like
        Thickness of the AEM layers
    lith_data: pandas DataFrame including ['From', 'To', 'Code']
        Lithology logs

    Returns
    -------
    fraction : (n_layer, 2) ndarray, float
        columns of fine and coarse fractions
    """
    z_top = lith_data.From.values
    z_bottom = lith_data.To.values
    z = np.r_[z_top, z_bottom[-1]]
    code = lith_data.Code.values
    zmin = z_top.min()
    zmax = z_bottom.max()
    depth = np.r_[0.0, np.cumsum(hz)][:]
    z_aem_top = depth[:-1]
    z_aem_bottom = depth[1:]
    fraction = np.zeros((len(hz), 2), dtype=float) * np.nan
    hz_lith = z_bottom - z_top
    inds_active = np.logical_and(depth[: len(hz)] >= zmin, depth[: len(hz)] <= zmax)
    hz_active = hz[inds_active]

    mesh_1d_aem = TensorMesh([hz_active])
    mesh_1d_lith = TensorMesh([hz_lith])
    P = volume_average(mesh_1d_lith, mesh_1d_aem)
    coarse_fraction_active = P @ lith_data["Code"].values
    fraction[inds_active, 1] = coarse_fraction_active
    fraction[inds_active, 0] = 1 - coarse_fraction_active
    return fraction


def generate_water_level_map(
    water_level_df: pd.DataFrame,
    em_data: EMDataset,
    dx=500,
    dy=500,
    max_distance=1000,
    k_nearest_points=200,
    water_level_contour_df=None,
    constant_water_level=30.0,
):
    lx = em_data.xy[:, 0].max() - em_data.xy[:, 0].min()
    ly = em_data.xy[:, 1].max() - em_data.xy[:, 1].min()
    x_pad = lx * 0.1
    y_pad = ly * 0.1

    if water_level_df.shape[0] == 0:
        wse_em = np.ones(em_data.num_soundings, dtype=float) * constant_water_level
        f_int_wse = NearestNDInterpolator(em_data.xy, wse_em)
    else:
        xy_wse = water_level_df[["UTMX", "UTMY"]].values
        wse = water_level_df["GSE_WSE"].values

        if water_level_contour_df is not None:
            xy_wse_contour = water_level_contour_df[["UTMX", "UTMY"]].values
            xy_wse = np.vstack((xy_wse, xy_wse_contour))
            wse = np.r_[wse, water_level_contour_df["GSE_WSE"].values]

        # linear interpolation
        f_int_wse = NearestNDInterpolator(xy_wse, wse)
        wse_em = f_int_wse(em_data.xy)

    x, y, wse_idw = inverse_distance_interpolation(
        em_data.xy,
        wse_em,
        dx=dx,
        dy=dy,
        max_distance=max_distance,
        k_nearest_points=k_nearest_points,
        power=0,
        x_pad=x_pad,
        y_pad=y_pad,
    )

    X, Y = np.meshgrid(x, y)

    df_water_table_grid = pd.DataFrame(
        data=np.c_[X.flatten(), Y.flatten(), wse_idw.data.flatten()],
        columns=["x", "y", "water_table"],
    )

    return dict(
        x=x,
        y=y,
        X=X,
        Y=X,
        wse_idw=wse_idw,
        wse_em=wse_em,
        df=df_water_table_grid,
        nn_interpolator=f_int_wse,
    )


def compute_colocations(
    distance_threshold: int, em_data: EMDataset, df_lithology: pd.DataFrame
):
    lithology_group = df_lithology.groupby("WELL_ID")
    df_lithology_collar = lithology_group[["UTMX", "UTMY"]].mean()
    xy_lithology = df_lithology_collar[["UTMX", "UTMY"]].values

    # find the well locations that are within the distance_threshold of any em suruvey location
    xy_lithology_colocated, inds_lithology_colocated = find_locations_in_distance(
        em_data.xy, xy_lithology, distance=distance_threshold
    )

    # use each well location to lookup the cloest em survey location
    d_aem_colocated, inds_aem_colocated = find_closest_locations(
        xy_lithology_colocated, em_data.xy
    )

    # get the subset of co-located survey locations
    xy_aem_colocated = em_data.xy[inds_aem_colocated, :]

    # get the subset of co-located wells
    df_lithology_collar_colocated = df_lithology_collar.loc[
        df_lithology_collar.index[inds_lithology_colocated]
    ]

    # get names of the colocated wells
    well_names_colocated = df_lithology_collar_colocated.index.to_list()

    # there should always be the same number of co-ocated wells and survey locations
    assert inds_aem_colocated.size == inds_lithology_colocated.size

    n_colocated = inds_aem_colocated.size
    mean_separation_distance = d_aem_colocated.mean()
    return dict(
        n_colocated=n_colocated,
        mean_separation_distance=mean_separation_distance,
        xy_em=xy_aem_colocated,
        lithology_collar=df_lithology_collar_colocated,
        well_names=well_names_colocated,
        inds_em=inds_aem_colocated,
        inds_lithology=inds_lithology_colocated,
    )



import json
from collections import namedtuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


class EMDataset:
    filename: str
    df: pd.DataFrame

    def __init__(
        self, csv_filename: str, csv_filename_thickness: str, spatial_unit="m"
    ):
        self.filename = csv_filename
        self.filename_thickness = csv_filename_thickness
        self.df = pd.read_csv(
            self.filename, dtype={"LINE_NO": "O", "RECORD": "Int32"}
        ).sort_values("RECORD")
        self.df_thickness = pd.read_csv(self.filename_thickness)
        self.spatial_unit = spatial_unit
        if spatial_unit == "ft":
            self.spatial_conversion = 1.0 / 0.3048
        elif spatial_unit == "m":
            self.spatial_conversion = 1.0

    @property
    def header(self):
        return list(self.df.columns)

    @property
    def line(self):
        return self.df["LINE_NO"].values

    @property
    def timestamps(self):
        return self.df["RECORD"].values.to_numpy(dtype=int)

    @property
    def topography(self):
        return (
            self.df[["UTMX", "UTMY", "ELEVATION"]].values[:, :]
            * self.spatial_conversion
        )

    @property
    def depth_of_investigtaion(self):
        return self.df["DOI_STANDARD"].values * self.spatial_conversion

    @property
    def hz(self):
        hz = np.array(json.loads(self.df_thickness.THICKNESS[0]))
        return np.r_[hz, hz[-1]] * self.spatial_conversion

    @property
    def depth(self):
        return np.cumsum(np.r_[0, self.hz])

    @property
    def resistivity(self):
        if getattr(self, "_resistivity", None) is None:
            resistivity = []
            for string in self.df["MEASUREMENTS"]:
                resistivity.append(json.loads(string)["RHO"])
            self._resistivity = np.vstack(resistivity)
        return self._resistivity

    @property
    def minmax(self):
        flat = self.resistivity.flatten()
        return np.min(flat), np.max(flat)

    @property
    def num_soundings(self):
        return self.df.shape[0]

    @property
    def xy(self):
        return self.df[["UTMX", "UTMY"]].values * self.spatial_conversion

    @property
    def lines_xy(self):
        lines = []
        em_lines = dict()
        for g, data in self.df[["LINE_NO", "RECORD", "UTMX", "UTMY"]].groupby(
            "LINE_NO"
        ):
            lines.append(g)
            em_lines[g] = (
                data[["UTMX", "UTMY", "RECORD"]].values * self.spatial_conversion
            )
        return lines, em_lines

    @property
    def num_layers(self):
        return self.hz.size

    def get_resistivity_by_line(self, line_number: int):
        records = self.df[self.df["LINE_NO"] == line_number][["UTMX", "UTMY"]]
        inds_line = self.line == line_number
        xy = self.xy[inds_line, :]
        rho = self.resistivity[inds_line, :]
        doi = self.depth_of_investigtaion[inds_line]
        elevation = self.topography[inds_line, 2]
        start_x = xy[0, 0]
        end_x = xy[-1, 0]
        if start_x > end_x:
            rho = rho[::-1, :]
            xy = xy[::-1, :]

        delta = np.concatenate(
            [[0], (np.diff(xy[:, 0]) ** 2 + np.diff(xy[:, 1]) ** 2) ** 0.5]
        )
        return rho, delta, xy, elevation, doi

    def get_binned_resistivity_by_line(
        self, line_number: int, n_bins: int, maximum_depth: float
    ):
        rho, delta, xy, elevation, doi = self.get_resistivity_by_line(line_number)
        distance = np.cumsum(delta)
        d_max = distance.max()
        distance_bins = np.linspace(0, d_max, n_bins + 1)
        inds = np.digitize(distance, distance_bins)
        depth_all = np.cumsum(np.r_[0.0, self.hz[:-1]])
        inds_above = depth_all < maximum_depth
        n_layer = inds_above.sum()
        rho_bins = np.zeros((n_bins, n_layer), dtype=float) * np.nan
        doi_bins = np.zeros(n_bins, dtype=float) * np.nan
        elevation_bins = np.zeros(n_bins, dtype=float) * np.nan
        for ii in range(n_bins):
            inds_tmp = inds == ii
            if inds_tmp.sum() > 0:
                rho_bins[ii, :] = rho[inds_tmp, :n_layer].mean(axis=0)
                doi_bins[ii] = doi[inds_tmp].mean(axis=0)
                elevation_bins[ii] = elevation[inds_tmp].mean(axis=0)
        distance_bins_center = (distance_bins[:-1] + distance_bins[1:]) * 0.5
        f_doi = interp1d(
            distance_bins_center[~np.isnan(doi_bins)],
            doi_bins[~np.isnan(doi_bins)],
            fill_value="extrapolate",
        )
        doi_bins_filled = f_doi(distance_bins_center)
        f_elevation = interp1d(
            distance_bins_center[~np.isnan(elevation_bins)],
            elevation_bins[~np.isnan(elevation_bins)],
            fill_value="extrapolate",
        )
        elevation_bins_filled = f_elevation(distance_bins_center)
        depth = np.cumsum(np.r_[0.0, self.hz[:n_layer]])
        return (
            distance,
            distance_bins,
            depth,
            rho_bins,
            elevation_bins_filled,
            doi_bins_filled,
        )



import numpy as np
import numpy.ma as ma
import skfmm
from discretize import TensorMesh
from SimPEG import utils


def classify_cf(cf, threshold):
    """
    Classify coarse fraction values into two categories: low-K and high-K unit

    Parameters
    ----------

    cf: array_like, float
        coarse
    threshold: float
        a threshold value to classify coarse fraction values into low-K and high-K unit
    Returns
    -------

    resistivity_for_lithology: array_like
        bootstrapped resistivity values for each lithology, size of the matrix is (n_bootstrap, n_lithology)
    """
    binary_model = np.ones_like(cf)
    binary_model[cf < threshold] = 0.0
    return binary_model


# ==================================== RECHARGE METRICS ====================================#


def get_ready_for_fmm(mesh, xyz_water):
    """
    Extend mesh for running fast marching and find indices for the source.

    Parameters
    ----------

    mesh: object
        discretize TensorMesh object.
    xyz_water: array_like, (*,3)
        locations of water table.

    Returns
    -------

    mesh_fmm: array_like
        discretize TensorMesh object.
    inds_source: array_like, bool
        indicies of the source.
    """
    dx = mesh.hx.min()
    dy = mesh.hy.min()
    dz = mesh.hz.min()

    hx = np.r_[mesh.hx[0], mesh.hx, mesh.hx[-1]]
    hy = np.r_[mesh.hy[0], mesh.hy, mesh.hy[-1]]
    hz = np.r_[mesh.hz, mesh.hz[-1]]
    x0 = [mesh.x0[0] - mesh.hx[0], mesh.x0[1] - mesh.hx[1], mesh.x0[2]]
    dz = hz.min()
    mesh_fmm = TensorMesh([hx, hy, hz], x0=x0)
    xyz_wse_up = np.c_[xyz_water[:, :2], -xyz_water[:, 2]]
    xyz_wse_down = np.c_[xyz_water[:, :2], -xyz_water[:, 2] - dz]
    inds_below_wt_up = utils.surface2ind_topo(mesh_fmm, xyz_wse_up)
    inds_below_wt_down = ~utils.surface2ind_topo(mesh_fmm, xyz_wse_down)
    inds_source = inds_below_wt_down & inds_below_wt_up
    return mesh_fmm, inds_source


def run_fmm(binary_values, mesh_fmm, inds_source, vel_ratio=None):
    """
    Run fast marching to calculate distance from surface to water table.

    Parameters
    ----------

    binary_values: array_like (*,)

    mesh_fmm: object
        discretize TensorMesh object.
    xyz_water: array_like, (*,3)
        locations of water table.
    inds_source:
        indicies of the source.
    vel_ratio: None or float
        ratio of the infiltration rate between low-K and high-K unit.

    Returns
    -------

    shortest_distance: array_like, (*,*)
        2D map of the shortest distance.
    """

    nx, ny, nz = mesh_fmm.vnC
    values_tmp = np.random.randint(low=0, high=2, size=nx * ny * nz).reshape(
        (nx, ny, nz)
    )
    values_tmp[1:-1, 1:-1, :-1] = binary_values.reshape(
        (nx - 2, ny - 2, nz - 1), order="F"
    )
    values_tmp[:, :, -1] = np.ones((nx, ny))
    values_tmp = utils.mkvc(values_tmp)
    dx = mesh_fmm.hx[0]
    dy = mesh_fmm.hy[0]
    dz = mesh_fmm.hz[0]
    # source term
    phi = np.ones(mesh_fmm.nC)
    # speed (assume 1 m/s)
    speed = np.ones(mesh_fmm.nC) * 1
    inds_noflow = values_tmp == 0.0
    # mask indicies (I assumed that clay unit as a barrier)
    if vel_ratio is None:
        mask = np.zeros(mesh_fmm.nC, dtype=bool)
        mask[inds_noflow] = True
        phi = np.ma.MaskedArray(phi, mask)
        phi[inds_source] = -1
        d = skfmm.distance(
            phi.reshape(mesh_fmm.vnC, order="F").transpose((1, 0, 2)),
            dx=[dx, dy, dz],
        )
        shortest_distance = d.transpose((1, 0, 2))[1:-1, 1:-1, -1].flatten(order="F")

    else:
        mask = np.zeros(mesh_fmm.nC, dtype=bool)
        inds_noflow = values_tmp == 0.0
        if vel_ratio >= 1.0:
            raise Exception(
                "input vel_ratio is {:.1e}, this should be less than 1".format(
                    vel_ratio
                )
            )
        speed[inds_noflow] = vel_ratio
        phi = np.ma.MaskedArray(phi, mask)
        phi[inds_source] = -1
        # travel time (since we assumed 1m/s, t is same as distance)
        # t and d will be different when you assign variable velocity
        t = skfmm.travel_time(
            phi.reshape(mesh_fmm.vnC, order="F").transpose((1, 0, 2)),
            speed.reshape(mesh_fmm.vnC, order="F").transpose((1, 0, 2)),
            dx=[dx, dy, dz],
            order=1,
        )
        shortest_distance = t.transpose((1, 0, 2))[1:-1, 1:-1, -1].flatten(order="F")
    return shortest_distance


def get_distance_to_first_lowK(values, inds_below, z):
    inds_lowK = np.logical_or(values == 0, inds_below)
    z_inds = np.ones(inds_lowK.shape[0], dtype=int) * -1
    for ii in range(inds_lowK.shape[0]):
        tmp = np.argwhere(inds_lowK[ii, :])
        if tmp.size > 0:
            z_inds[ii] = tmp.max()
    shortest_distance = z[z_inds]
    return -shortest_distance



import sys
import os
import setuptools

sys.path[0:0] = ['emrecharge']
from version import __version__

def readme():
    if os.path.exists("README.md"):
        with open("README.md") as file:
            return file.read()
    return ""

setuptools.setup(
    name="emrecharge",
    version=__version__,
    description="Tools for EM Groundwater Recharge",
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Hydrology",
        "Intended Audience :: Science/Research",
    ],
    packages=setuptools.find_packages(include=['emrecharge']),
    python_requires='>=3.6',
    install_requires=[
        "rasterio",
        "shapely",
        "pandas",
        "numpy",
        "scipy",
        "verde",
        "scikit-fmm",
        "discretize",
        "SimPEG",
        "geopandas",
        "matplotlib",
        "properties[math, image]"
    ],
)


import geopandas as gpd
import numpy as np
import rasterio
from rasterio.transform import Affine
from shapely.geometry import LineString, Polygon


# Get one of the contours from the plot.
def create_gpd_dataframe_from_contours(
    cs, input_crs, output_crs, spatial_conversion, dx, dy
):
    geometry = []
    index = []
    i_count = 0
    for cnt in cs.collections:
        n = len(cnt.get_paths())
        for ii in range(n):
            vs = cnt.get_paths()[ii].vertices
            vs = np.c_[vs[:, 0] - dx / 2, vs[:, 1] - dy / 2] / spatial_conversion
            polygon_geom = LineString(vs)
            geometry.append(polygon_geom)
            index.append(i_count)
            i_count += 1
    polyline = gpd.GeoDataFrame(index=index, crs=input_crs, geometry=geometry)
    polyline = polyline.to_crs(crs=output_crs)
    return polyline


def export_to_tif(data, dx, dy, xmin, ymin, crs, fname):
    transform = Affine.translation(xmin - dx / 2, ymin - dy / 2) * Affine.scale(dx, dy)
    tmp = data.copy()
    new_dataset = rasterio.open(
        fname,
        "w",
        driver="GTiff",
        height=tmp.shape[0],
        width=tmp.shape[1],
        count=1,
        dtype=tmp.dtype,
        crs=crs,
        transform=transform,
    )
    new_dataset.write(tmp[:, :], 1)
    new_dataset.close()
